/***************************************************************************\
  "parser.h" - parser
  Rev 1; 10 Oct 2002
  Copyright (C) 2002  Kevin P. Albrecht

  Permission is hereby granted by Kevin P. Albrecht (the "Author"), free of
  charge, to any person obtaining a copy of this software and associated
  documentation files (the "Software"), to deal in the Software without
  restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the
  Software, and to permit persons to whom the Software is furnished to do
  so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included
  in all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
  THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
  AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

  Except as contained in this notice, the name of the Author shall not be
  used in advertising or otherwise to promote the sale, use or other
  dealings in this Software without prior written authorization from the
  Author.
\***************************************************************************/

#ifndef PARSER_H
#define PARSER_H

/////////////////////////
// DEFINES //////////////
/////////////////////////

#define PARSER_STRING_LEN 256

/////////////////////////
// INCLUDES /////////////
/////////////////////////

#include "table.h"
#include "memory.h"
#include "lexer.h"

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

Word *thisword, *lastword;

// Private //
void parser_verify ();
void parser_getword ();

// Public //
void parser_parse ();

// Private - Parse Functions //
Symbol *parse_expr ();
void parse_SET ();
void parse_PRINT ();
void parse_get ();
void parse_math ();
void parse_LABEL ();
void parse_GOTO ();
void parse_jump ();

////////////////////////
// DEFINITIONS /////////
////////////////////////

void parser_verify ()
{
  if (thisword == NULL)
  {
    ohno (true, "invalid EOF after (word:meaning)", lastword->string, WordCodeStrings[lastword->code]);
  }
}

void parser_getword ()
{
  //BUG// word_del (lastword);

  lastword = thisword;
  thisword = lexer_next ();
}

void parser_parse ()
{
  Symbol *s;

  // var_51: the number zero //
  s = tab_enter (gTable, "var_51");
  s->value = alloc_n (0);
  s->type = tNumber;

  // var_52: the number one //
  s = tab_enter (gTable, "var_52");
  s->value = alloc_n (1);
  s->type = tNumber;

  // var_53: the newline string //
  s = tab_enter (gTable, "var_53");
  s->value = k_clone ("\n");
  s->type = tString;

  lastword = NULL;
  thisword = NULL;

  parser_getword ();

  while (thisword != NULL)
  {
    switch (thisword->code)
    {
      case wcSET:
        parse_SET ();
        break;
      case wcPRINT:
        parse_PRINT (); 
        break;
      case wcGETN:
      case wcGETS:
        parse_get ();
        break;
      case wcADD:
      case wcSUB:
      case wcMUL:
      case wcDIV:
        parse_math ();
        break;
      case wcLABEL: 
        parse_LABEL (); 
        break;
      case wcGOTO:
        parse_GOTO ();  
        break;
      case wcJLT:
      case wcJGT:
      case wcJE:
      case wcJNE:
        parse_jump ();
        break;
      default:
        ohno (true, "incorrect instruction encountered (word:meaning)", thisword->string, WordCodeStrings[thisword->code]);
        break;
    }

    parser_getword ();
  }
}

Symbol *parse_expr ()
{
  Symbol *temp;
  char buf[PARSER_STRING_LEN];
  int pos;

  if (thisword->code == wcSTROPEN)
  {
    parser_getword ();
    parser_verify ();

    for (pos = 0; thisword->code != wcSTRCLOSE; pos++)
    {
      buf[pos] = (char) thisword->value;

      parser_getword ();
      parser_verify ();
    }

    buf[pos] = '\0';

    temp = tab_enter (gTable, buf);
    temp->type = tString;
    temp->value = k_clone (buf);
  }
  else if (thisword->code == wcNUMBER)
  {
    parser_getword ();
    parser_verify ();

    temp = tab_enter (gTable, k_int (thisword->value));
    temp->type = tNumber;
    temp->value = alloc_n (thisword->value);
  }
  else
  {
    temp = tab_search (gTable, thisword->variable);

    if (temp == NULL)
    {
      ohno (true, "variable in the expression not defined (word)", thisword->string, NULL);
    }
  }

  return temp;
}

void parse_SET ()
{
  Symbol *a, *b;

  parser_getword ();
  parser_verify ();

  a = tab_enter (gTable, thisword->variable);

  parser_getword ();
  parser_verify ();

  b = parse_expr ();

  memory_write (wcSET, a, b, NULL);
}

void parse_PRINT ()
{
  Symbol *a;

  parser_getword ();
  parser_verify ();

  a = parse_expr ();

  memory_write (wcPRINT, a, NULL, NULL);
}

void parse_get ()
{
  WordCode wc;
  Symbol *a;

  wc = thisword->code;

  parser_getword ();
  parser_verify ();

  a = tab_enter (gTable, thisword->variable);

  memory_write (wc, a, NULL, NULL);
}

void parse_math ()
{
  WordCode wc;
  Symbol *a, *b;

  wc = thisword->code;

  parser_getword ();
  parser_verify ();

  a = tab_enter (gTable, thisword->variable);

  parser_getword ();
  parser_verify ();

  b = parse_expr ();

  memory_write (wc, a, b, NULL);
}

void parse_LABEL ()
{
  Symbol *a;

  parser_getword ();
  parser_verify ();

  a = tab_search (gTable, thisword->variable);
  
  if (a)
  {
    if (a->type == tNull) // "a" was already defined in a jump
    {
      a->type = tLabel;
      a->value = alloc_n (ip);
    }
    else // "a" was previously defined in a "label"
    {
      ohno (true, "cannot redefine a label (word)", thisword->string, NULL);
    }
  }
  else // "a" is not yet defined
  {
    a = tab_enter (gTable, thisword->variable);
    a->type = tLabel;
    a->value = alloc_n (ip);
  }

  memory_write (wcLABEL, a, NULL, NULL);
}

void parse_GOTO ()
{
  ohno (true, "\"goto\" not implemented yet (word)", thisword->string, NULL);
}

void parse_jump ()
{
  WordCode wc;
  Symbol *a, *b, *c;

  wc = thisword->code;

  parser_getword ();
  parser_verify ();

  a = tab_enter (gTable, thisword->variable);

  parser_getword ();
  parser_verify ();

  b = tab_enter (gTable, thisword->variable);

  parser_getword ();
  parser_verify ();

  c = tab_enter (gTable, thisword->variable);

  memory_write (wc, a, b, c);
}

#endif
