/***************************************************************************\
  "table.h" - symbol table
  Rev 1; 10 Oct 2002
  Copyright (C) 2002  Kevin P. Albrecht

  Permission is hereby granted by Kevin P. Albrecht (the "Author"), free of
  charge, to any person obtaining a copy of this software and associated
  documentation files (the "Software"), to deal in the Software without
  restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the
  Software, and to permit persons to whom the Software is furnished to do
  so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included
  in all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
  THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
  AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

  Except as contained in this notice, the name of the Author shall not be
  used in advertising or otherwise to promote the sale, use or other
  dealings in this Software without prior written authorization from the
  Author.
\***************************************************************************/

#ifndef TABLE_H
#define TABLE_H

/////////////////////////
// DEFINES //////////////
/////////////////////////

// This value (known as "m") should be a prime number so that
//   if "n" is the average number of symbols in the symbol table,
//   "n / m" is approximately <= 2.  See (Aho pp433-437).
// According to Grune, a good "m" is 1008 (MCD 2.1.11.3)
#define TABLE_HASH_ARRAY_SIZE 17

// This value (known as "alpha") is multiplied by each character
//   in the symbol's key before they are added together.
//   According to (Aho p437), alpha=1 is the least efficient
//   and alpha=65599 is best for 32-bit integers.
// According to Grune, a good "alpha" is 613 (MCD 2.1.11.3)
#define TABLE_HASH_ALPHA 65599

#define TABLE_MAX_SYMBOLS 256

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //

// Local //
#include "k_string.h"
#include "global.h"

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

typedef enum
{
  tNull, tLabel, tString, tNumber
} Type;

char *TypeStrings[] =
{
  "null", "label", "string", "number"
};

// Symbol //
typedef struct Symbol_tag
{
  Type type;

  void *value; // the value held in the symbol:
               // TYPE         VALUE          MEANING
               // ----         -----          -------
               // tNull        (not defined)  (not defined)
               // tLabel       int*           instruction index
               // tString      char*          string
               // tNumber      int*           integer

  struct Symbol_tag *next;
  char *key;
} Symbol;

Symbol *sym_new (char *aKey);
void sym_del (Symbol *symbol);
void sym_set (Symbol *symbol, Type aType, void *aValue);

// Bucket //
typedef struct
{
  Symbol *head;
  Symbol *tail;
} Bucket;

void buc_init (Bucket *bucket);
Symbol *buc_search (Bucket *bucket, char *aKey);
Symbol *buc_enter (Bucket *bucket, char *aKey);

// Table //
typedef struct
{
  Bucket **array;
} Table;

Table *tab_new ();
void tab_del (Table *table);
Symbol *tab_search (Table *table, char *aKey);
Symbol *tab_enter (Table *table, char *aKey);

// Hash //
int hash (char *as);

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

//// sym_new
Symbol *sym_new (char *aKey)
{
  Symbol *temp = (Symbol *) malloc (sizeof (Symbol));

  temp->type = tNull;
  temp->value = NULL;

  temp->next = NULL;
  temp->key = k_new (strlen (aKey)); k_copy (temp->key, aKey);

  return temp;
}

//// sym_del
void sym_del (Symbol *symbol)
{
  // symbol may be NULL //
  if (!symbol) return;

  if (symbol->value) free (symbol->value);

  k_del (symbol->key);
  free (symbol);
}

//// sym_set
void sym_set (Symbol *symbol, Type aType, void *aValue)
{
  symbol->type = aType;

  // given value is NULL, so set the value as the default for each type //
  if (aValue == NULL)
  {
    // determine the type of the declaration //
    if (aType == tNull) symbol->value = NULL;
    else if (aType == tLabel) symbol->value = alloc_n (0);
    else if (aType == tString) symbol->value = k_clone ("");
    else if (aType == tNumber) symbol->value = alloc_n (0);
  }
  else
  {
    if (aType == tNull) symbol->value = NULL;
    else if (aType == tLabel) symbol->value = alloc_n (*(int *)aValue);
    else if (aType == tString) symbol->value = k_clone (aValue);
    else if (aType == tNumber) symbol->value = alloc_n (*(int *)aValue);
  }

}

//// buc_init
void buc_init (Bucket *bucket)
{
  bucket->head = NULL;
  bucket->tail = NULL;
}

//// buc_search
//// Returns pointer to the Symbol with the matching
//// symbol key, or NULL if not found
Symbol *buc_search (Bucket *bucket, char *aKey)
{
  Symbol *temp = bucket->head;

  // Loop to find the matching symbol key //
  while (temp)
  {
    if (k_same (temp->key, aKey))
    {
      return temp;
    }
    else
    {
      temp = temp->next;
    }
  }

  // Not found; return NULL //
  return NULL;
}

//// buc_enter
//// Returns pointer to the Symbol; If not found, it will be created
Symbol *buc_enter (Bucket *bucket, char *aKey)
{
  Symbol *temp;

  temp = buc_search (bucket, aKey);

  // Not found; create and enter it //
  if (temp == NULL)
  {
    temp = sym_new (aKey);

    // First element //
    if (bucket->head == NULL)
    {
      bucket->head = bucket->tail = temp;
    }
    // Other elements //
    else
    {
      bucket->tail->next = temp;
      bucket->tail = temp;
    }
  }

  // Found //
  else
  {
  }

  // Return //
  return temp;
}

//// tab_new
Table *tab_new ()
{
  int i;
  Table *temp = (Table *) malloc (sizeof (Table));

  // Initialize the hash array to NULL //
  temp->array = (Bucket **) malloc (sizeof (Bucket *) * TABLE_HASH_ARRAY_SIZE);
  for (i = 0; i < TABLE_HASH_ARRAY_SIZE; i++)
  {
    temp->array[i] = NULL;
  }

  return temp;
}

//// tab_del
void tab_del (Table *table)
{
  int i;

  // free each element in array //
  for (i = 0; i < TABLE_HASH_ARRAY_SIZE; i++)
  {
    free (table->array[i]);
  }
  free (table->array);

  free (table);
}

//// tab_search
//// Returns pointer to the node, or NULL if not found
Symbol *tab_search (Table *table, char *aKey)
{
  Bucket *pBucket;
  Symbol *pSymbol;

  pBucket = table->array[hash (aKey)];

  // There is no Bucket at that hash; Not found //
  if (pBucket == NULL)
  {
    return NULL;
  }
  // Bucket found at that hash //
  else
  {
    pSymbol = buc_search (pBucket, aKey);
    return pSymbol;
  }
}

//// tab_enter
//// Returns pointer to the node; If not found, it will be created
Symbol *tab_enter (Table *table, char *aKey)
{
  int nHash;
  Bucket *pBucket;
  Symbol *pSymbol;

  nHash = hash (aKey);
  pBucket = table->array[nHash];

  // There is no Bucket at that hash; Create Bucket //
  if (pBucket == NULL)
  {
    pBucket = (Bucket *) malloc (sizeof (Bucket));
    buc_init (pBucket);
    table->array[nHash] = pBucket;
  }

  // Create Symbol //
  pSymbol = buc_enter (pBucket, aKey);

  return pSymbol;
}

//// hash
//// hash function for the "Alpha" algorithm (Aho p435)
int hash (char *as)
{
  int i;
  int len;
  int total;

  len = strlen (as);
  total = 0;

  for (i = 0; i < len; i++)
  {
    total += as[i] * TABLE_HASH_ALPHA;
  }

  return (total % TABLE_HASH_ARRAY_SIZE);
}

#endif

